
/* SCCSID = %W% %E% */
/****************************************************************************
 *                                                                          *
 * Copyright (c) IBM Corporation 1994 - 1997.                               *
 *                                                                          *
 * The following IBM OS/2 source code is provided to you solely for the     *
 * the purpose of assisting you in your development of OS/2 device drivers. *
 * You may use this code in accordance with the IBM License Agreement       *
 * provided in the IBM Device Driver Source Kit for OS/2.                   *
 *                                                                          *
 ****************************************************************************/
/**@internal %W%
 * @notes
 * @version %I%
 * @context Unless otherwise noted, all interfaces are Ring-0, 16-bit,
 *  <stack context>.
 * @history
 *
 */
#define INCL_NOPMAPI
#define INCL_DOSERRORS            // for ERROR_INVALID_FUNCTION
#include <os2.h>
#include <os2me.h>
#include <audio.h>                // for #define MIDI
#include <string.h>
#include <include.h>
#include <devhelp.h>

#include "wavestrm.hpp"
#include "audiohw.hpp"
#include "waudio.hpp"
#include "memutil.h"
#include <ossidc.h>
#include <dbgos2.h>
#include "mixer.h"
#include "convert.h"

#ifndef min
#define min(a,b) (a>b) ? b : a
#endif


//******************************************************************************
//  _vRealignBuffer
//  called just after a wave stream pause on a playback.
//  Gets the end position of the stream when paused and a pointer to a
//  STREAMBUFFER. Basicly this function looks at the streambuffer and if
// there is any unplayed data in it it adjusts the bufpos counter.
// the donepos counter is ALWAYS set to zero. It will return 0 if all
// the data has been played and 1 if there is still some data left.
//******************************************************************************
USHORT WAVESTREAM::_vRealignBuffer(ULONG FAR *bytesinc, PSTREAMBUFFER pbuffer)
{
    // if none of the data in this stream buffer has been consumed
    if (!*bytesinc) {
      	pbuffer->ulDonepos = 0;
      	pbuffer->ulBuffpos = 0;
      	return 1;
    }

    pbuffer->ulDonepos += *bytesinc;
    pbuffer->ulBuffpos  = pbuffer->ulDonepos;
    *bytesinc           = 0;
    if(pbuffer->ulDonepos >= pbuffer->ulBuffsz) {
	    //calc position in next buffer
	    *bytesinc = pbuffer->ulDonepos - pbuffer->ulBuffsz;
	    return 0; //all of the buffer has been consumed
    }
    return 1;
}
//******************************************************************************
// _vRealignPausedBuffers(void)
// when a stream is paused we need to "realign" the data in the audio buffer
// with reality. On playback, not all the data in the audio buffer has been
// consumed. Likewise on a capture, not all the good data in the audio buffer
// has been copied out. After receiving the DDCMDCONTROL Pause we will call
// this function to line the MMPM buffers back up.
// there are 2 cases here: first one is the case of a capture stream.
// for a capture stream we simply read any data that is still in the audio
// buffer into a MMPM buffer.
// for a playback stream things are not so straight forward.
// first check the STREAMBUFFER on pHead to see if any of it's data is in the
// audio buffer and not consumed, if yes back up the ulBuffpos in the
// STREAMBUFFER. Next check any STREAMBUFFERS on pdone starting with the last
// one. (the one on the tail of the queue) If necessary back up the ulBuffpos
// and put the STREAMBUFFER on the Head queue.
//******************************************************************************
void WAVESTREAM::_vRealignPausedBuffers(ULONG endpos)
{
    PSTREAMBUFFER ptempbuff;

    switch (ulStreamType & STREAM_WRITE) {
    case STREAM_READ:
	    //SvL: Don't get the lastest recording data as a read command
        //     would now restart recording (as it's stopped)
        //     Just return what we've got or push the buffer on the inprocess queue
	    ptempbuff = (PSTREAMBUFFER)qhDone.Head();
	    if(ptempbuff) {
		    if(ptempbuff->ulBuffpos) {//if we recorded anything into this buffer, then return it now
    			ReturnBuffer();
			    return;
		    }
		    ptempbuff->ulBuffpos = 0;
		    ptempbuff->ulDonepos = 0;
		    qhInProcess.PushOnHead(qhDone.PopHead());
	    }
      	break;

   case STREAM_WRITE:
   {
        PQUEUEHEAD pTempHead = new QUEUEHEAD;
        ULONG bytesinc;
        USHORT usRC;

	    bytesinc = endpos - _ulBytesProcessed;
      	bytesinc &= 0xFFFFFFFC; //keep it on a dword boundary

      	// if there are bufferes on the done queue, pop them off the head and
      	// push them on the head of qhTempHead.  This will reorder them so
      	// that the more recently used ones will be in the front of the queue.
      	// Pass them all to _vRealignBuffer. If the rc from _vRealignBuffer is
      	// 0 then there is no unprocessed data in the buffer (it is ready to
      	// be returned) so put it on the Tail of the done queue.
      	// If the rc is 1 then put it on the head of the InProcess queue.

      	while (qhDone.IsElements()) {
         	pTempHead->PushOnTail(qhDone.PopHead());
      	} /* endwhile */

      	while (qhInProcess.IsElements()) {
         	pTempHead->PushOnTail(qhInProcess.PopHead());
      	} /* endwhile */

      	while(pTempHead->IsElements()) {
         	usRC = _vRealignBuffer(&bytesinc, (PSTREAMBUFFER)pTempHead->Head());
         	if (usRC) {
           		qhInProcess.PushOnTail(pTempHead->PopHead());
		    }
         	else {
		    	qhDone.PushOnTail(pTempHead->PopHead());
		    }
      	} /* endwhile */
	    if(qhDone.IsElements())
    		ReturnBuffer();

       	delete pTempHead; // free the memory this ain't no Java here !!
      	break;
    }
    default:
      	break;
    } /* endswitch */
}
//******************************************************************************
// get ready to start streaming
// this requires the following:
// call Initbuffer in the audiobuffer object
// if this is a write stream call _vFillAudioBuf
//******************************************************************************
#pragma off (unreferenced)
void WAVESTREAM::AddBuffers(BOOL fFirst)
#pragma on (unreferenced)
{
    ULONG space, byteswritten;

    if (ulStreamType & STREAM_WRITE) {
	    if(!qhInProcess.Head() && !qhDone.Head()) {
    		//underrun: stop playback
//		    dprintf(("underrun: stop playback"));

            //save last known position
            _ulUnderrunBase = GetCurrentPos();
            
            //silence wave stream before stopping it (removes clicks)
            pahw->SetVolume(StreamId, getMixerStreamId(), 0);

            //and stop it
		    pahw->Stop(StreamId);
		    fUnderrun = TRUE;
		    return;
	    }
	    if(pahw->GetSpace(StreamId, &_configinfo, &space) == FALSE) {
            dprintf(("pahw->getSpace: No room!"));
            return;
        }
	    while(space) {
		    byteswritten = AddBuffer(space);
		    if(byteswritten == (ULONG)-1) break;
		    space -= byteswritten;
	    }
    }
}
//******************************************************************************
// write one buffer to the audio buffer
// the caller of this function MUST make sure it ok to write the audio buffer..
// _AudioBufWrite will not check if there is room in the audio buffer of if
// there are buffers on pHead... BEWARE
//******************************************************************************
ULONG WAVESTREAM::AddBuffer(ULONG space)
{
    PSTREAMBUFFER pTemp = (PSTREAMBUFFER)qhDone.Tail();
    ULONG pdataBuf;
    ULONG Buff_left, byteswritten;

    if(!pTemp || pTemp->ulBuffpos >= (pTemp->ulBuffsz & 0xFFFFFFFC)) {
	    pTemp = (PSTREAMBUFFER)qhInProcess.Head();
    }
    if(!pTemp) {
#ifdef DEBUG
	    dprintf(("AddBuffer: pTemp == NULL"));
#endif /* DEBUG */
	    return (ULONG)-1;
    }

    // get the buffer pointer and amount of data remaining
    pdataBuf  = (ULONG)pTemp->pBuffptr + pTemp->ulBuffpos;
    Buff_left = pTemp->ulBuffsz - pTemp->ulBuffpos;

    if(Buff_left) {
       // write the audio buffer
       Buff_left = min(Buff_left, space);
       if(pahw->Transfer(StreamId, &_configinfo, pdataBuf, Buff_left, &byteswritten) == FALSE) {
           dprintf(("AddBuffer: pahw->Transfer failed!!"));
           return (ULONG)-1;
       }
       if(byteswritten == 0) {
           return (ULONG)-1; //no more room
       }

       // update the buffer pos counter
       pTemp->ulBuffpos += byteswritten;
    } 
    else byteswritten = 0;

    if(pTemp == qhInProcess.Head()) {
    	qhDone.PushOnTail(qhInProcess.PopHead());
    }
    dprintf(("AddBuffer %lx size %d, bytes written %d", pdataBuf, (USHORT)Buff_left, (USHORT)byteswritten));
    return byteswritten;
}
//******************************************************************************
// Read data from the audio Buffer.
// Called at interrupt time to get the good data from the audiobuffer object.
//******************************************************************************
BOOL WAVESTREAM::_vReadAudioBuf(void)
{
    PSTREAMBUFFER pTemp = (PSTREAMBUFFER)qhInProcess.Head();
    ULONG pdataBuf;
    ULONG Buff_left, bytesread;

    if(!pTemp) 
        return FALSE;

    // get the buffer pointer and amount of data remaining
    pdataBuf  = (ULONG)pTemp->pBuffptr + pTemp->ulBuffpos;
    Buff_left = pTemp->ulBuffsz - pTemp->ulBuffpos;

    // read wave data
    if(pahw->Transfer(StreamId, &_configinfo, pdataBuf, Buff_left, &bytesread) == FALSE) {
       dprintf(("_vReadAudioBuf: pahw->Transfer failed!!"));
       return FALSE;
    }
    if(bytesread == 0) {
       return FALSE; //no more data
    }

    dprintf4(("_vReadAudioBuf %lx size %ld, bytes read %ld", pdataBuf, Buff_left, bytesread));

    // update the buffer pos counter
    pTemp->ulBuffpos  += bytesread;
    _ulBytesProcessed += bytesread;

    if(pTemp->ulBuffpos == pTemp->ulBuffsz) {
      	qhDone.PushOnTail(qhInProcess.PopHead());
	    ReturnBuffer();
   	    dprintf(("_vReadAudioBuf return buffer %lx size %ld, bytes read %ld position %ld", (ULONG)pTemp->pBuffptr, pTemp->ulBuffsz, bytesread, _ulBytesProcessed));
    }

    return TRUE;
}
//******************************************************************************
// called by the irq function in the hardware object when we get an interrupt
// first call _vUpdateProcessed() to update the dma amd audio buffer related
// stuff. Next if we have buffers on the primary queue try to read/write them
// to the audiobuffer. Look at the buffers on the done queue and see if they
// can be returned and finally process any events pending.
//******************************************************************************
void DBGCALLCONV WAVESTREAM::Process(void)
{
 PSTREAMBUFFER ptemp;
 ULONG         ulCurBytesProcessed = 0;
 ULONG         bytesinc;

#ifdef DEBUG
//    dprintf(("WAVESTREAM::Process"));
#endif

    switch (ulStreamType & STREAM_WRITE) {
    case STREAM_WRITE:
    {
        if(pahw->GetPosition(StreamId, &_configinfo, &ulCurBytesProcessed) == FALSE || 
           ulCurBytesProcessed == 0) 
        {//shouldn't happen
            DebugInt3();
            return;
        }

        bytesinc  = ulCurBytesProcessed - _ulBytesProcessed;
        dprintf3(("P: %lx %x", ulCurBytesProcessed, (USHORT)bytesinc));

        if(ulCurBytesProcessed < _ulBytesProcessed) 
        {
            ULONG ulCurBytesProcessed, ulStreamPos;

            if(OSS16_WaveGetPos(StreamId, &ulCurBytesProcessed) != OSSERR_SUCCESS) {
                ulCurBytesProcessed = 0;
            }

            ulStreamPos  = pahw->ConvertPositionInvert(ulCurBytesProcessed, &_configinfo);
            //round to sample boundary
            ulStreamPos &= ~(_configinfo.ulSampleSize - 1);

#ifdef DEBUG
            dprintf(("WARNING: Process: Current pos %ld pos %ld incr %d", ulCurBytesProcessed, _ulBytesProcessed, bytesinc));
            dprintf(("WARNING: hw pos %lx, streampos %lx", ulCurBytesProcessed, ulStreamPos));
#endif /* DEBUG */
        }
        _ulBytesProcessed = ulCurBytesProcessed;

        while(bytesinc) 
        {
      	  if(qhDone.IsElements()) {  // if there are buffers that have been completly written to the audio buffer
                                       // check the first one on the done queue
                                       // if it's data has been consumed by the hardware return it.
				                         // Bug #18 Sometimes MMPM gives us really small buffers. If we 
				                         // wait for the buffer to be completly consumed and there is a tiny buffer
				                         // on the queue after this one, the sound part could underrun before we get a 
				                         // chance to refill it. To fix this we will return the buffer if it will be  
 				                         // consumed before the next call to WAVESTREAM:Process(). 
				                         // rjj Fix for Bug #18 28082008
                ptemp = (PSTREAMBUFFER)qhDone.Head();
                ptemp->ulDonepos += bytesinc;
                bytesinc          = 0;
				  if ((ptemp->ulDonepos + _configinfo.ulBytesPerIRQ) >= ptemp->ulBuffsz)
                {
                    //calc position in next buffer
                    bytesinc = ptemp->ulDonepos - ptemp->ulBuffsz;
//                  dprintf2(("Process: Return buffer %lx size %d", ptemp->pBuffptr, ptemp->ulBuffsz));
                    ReturnBuffer();
		          }
      	  }
            else	break; //shouldn't happen
        }

        if (qhInProcess.IsElements()){
            AddBuffers(FALSE);
        }
        break;
    }
    case STREAM_READ:
	    while(_vReadAudioBuf());
	    break;
    default:
        break;
    } /* endswitch */

    ProcessEvents();
}
//******************************************************************************
//******************************************************************************
#pragma off (unreferenced)
ULONG WAVESTREAM::Write(PSTREAMBUF pbuf, ULONG uLength, BOOL fLooping)
#pragma on (unreferenced)
{
    PSTREAMBUFFER pStreamBuf = new STREAMBUFFER(uLength, pbuf);

    return Write(pStreamBuf);
}
//******************************************************************************
//******************************************************************************
ULONG WAVESTREAM::Write(PSTREAMBUFFER pStreamBuf)
{
    qhInProcess.PushOnTail((PQUEUEELEMENT)pStreamBuf);
//    dprintf2(("WAVESTREAM::Write: Push on tail %lx %ld", ((PSTREAMBUFFER)qhInProcess.Tail())->pBuffptr, ((PSTREAMBUFFER)qhInProcess.Tail())->ulBuffsz));
    if(fUnderrun) {
        PSTREAMBUFFER pTemp = (PSTREAMBUFFER)qhInProcess.Head();

//        dprintf(("Restart after underrun!"));
	    fUnderrun = FALSE;

        // configure the wave device
        // (NOTE: Must call this function; sample rate position is reset there)
        if(pahw->ConfigDev(StreamId, &_configinfo, pTemp->ulBuffsz) == FALSE) {
            dprintf(("ConfigDev failed!!"));
        }

        // prepare wave device for playback/recording
        if(pahw->Prepare(StreamId) == FALSE) {
            dprintf(("Prepare failed!!"));
        }
	    AddBuffers(TRUE);
	    if(ulStreamType == STREAM_WAVE_PLAY)
   		    MixerSetWaveVolume(getMixerStreamId(), StreamId, volume);
    }
    return 0;
}
//******************************************************************************
//******************************************************************************
ULONG WAVESTREAM::Read(PSTREAMBUF pbuf, unsigned uLength)
{
    qhInProcess.PushOnTail((PQUEUEELEMENT)new STREAMBUFFER(uLength, pbuf));
//    dprintf2(("WAVESTREAM::Read: Push on tail %lx %d", pbuf, uLength));
    return 0;
}
//******************************************************************************
// WAVESTREAM::GetCurrentTime(void)
// get current time will calculate the stream time in milliseconds based on
// NOW.... the open mpeg folks this is the greatest thing since my last
// pay raise!!
// but then again you know what those ring 3 programmers are like....
// the algorythum goes like this....
// bytes consumed / consume rate = seconds
// Note before calling BufferUpdate check to see if the stream is running.
// if it is not then call with flags of 0. This will prevent the audio buffer
// trying to get the latest consumption info from the hardware object. (dma or
// or pci as the case may be) Asking a hardware object that is not running for
// status information is just not a good idea.....
//******************************************************************************
ULONG WAVESTREAM::GetCurrentTime()
{
    ULONG Seconds, MilliSeconds, Overflow, Processed;

    if (ulStreamState == STREAM_STREAMING)  // if the stream is active
    {
   	    if (ulStreamType & STREAM_WRITE) 
        {
    		if(pahw->GetPosition(StreamId, &_configinfo, &Processed) == FALSE) 
            {
                DebugInt3();
                Processed = _ulBytesProcessed; //last known position
            }
	    }
	    else	Processed = _ulBytesProcessed;
    }
    else Processed = _ulBytesProcessed;

    //add position of underrun (if one occurred in the past)
    Processed += _ulUnderrunBase;

    // if we haven't processed anything then just return
    // _ulTimeBase
    if(Processed == 0)
        return(_ulTimeBase);

    Seconds = Processed / _configinfo.ulPCMConsumeRate;
    Overflow = Processed - (Seconds * _configinfo.ulPCMConsumeRate);
    MilliSeconds = (Overflow * 1000) / _configinfo.ulPCMConsumeRate;
    MilliSeconds += (Seconds * 1000);
    return(MilliSeconds + _ulTimeBase);
}
//******************************************************************************
//******************************************************************************
ULONG WAVESTREAM::GetCurrentPos(void)
{
    ULONG Processed;

    if (ulStreamState == STREAM_STREAMING)  // if the stream is active
    {
   	    if (ulStreamType & STREAM_WRITE) {
    		if(pahw->GetPosition(StreamId, &_configinfo, &Processed) == FALSE) 
            {
                DebugInt3();
                Processed = _ulBytesProcessed; //last known position
            }
	    }
	    else	Processed = _ulBytesProcessed;
    }
    else Processed = _ulBytesProcessed;

    //add position of underrun (if one occurred in the past)
    Processed += _ulUnderrunBase;
    return Processed;
}
//******************************************************************************
//******************************************************************************
ULONG WAVESTREAM::GetCurrentWritePos(void)
{
   ULONG writepos = 0;

   cli();
   PSTREAMBUFFER pTemp = (PSTREAMBUFFER)qhDone.Tail();

   if(!pTemp) {
       pTemp = (PSTREAMBUFFER)qhInProcess.Head();
   }
   if(pTemp) {
       writepos = pTemp->ulBuffpos;
   }
   sti();
   return writepos;
}
//******************************************************************************
// SetCurrentTime
// MMPM will send in the "starting stream time" as they see it.
// "our stream time" will always start at 0, so we save "their" time and
// add it to the elapsed time we calculate when we need to return time.
//******************************************************************************
void  WAVESTREAM::SetCurrentTime(ULONG time)
{
    _ulTimeBase = time;
}
//******************************************************************************
//******************************************************************************
ULONG WAVESTREAM::StartStream()
{
    PSTREAMBUFFER pTemp = (PSTREAMBUFFER)qhInProcess.Head();

#ifdef DEBUG
//    dprintf(("WAVESTREAM::StartStream"));
#endif

    if(ulStreamState == STREAM_STREAMING) {
        dprintf(("StartStream: already playing!!"));
        return NO_ERROR;
    }

    // configure the wave device
    // (NOTE: Must call this function; sample rate position is reset there)
    if(pahw->ConfigDev(StreamId, &_configinfo, (pTemp) ? pTemp->ulBuffsz : 0) == FALSE) {
        dprintf(("StartStream: ConfigDev failed!!"));
        goto fail;
    }

    // prepare wave device for playback/recording
    if(pahw->Prepare(StreamId) == FALSE) {
        dprintf(("StartStream: Prepare failed!!"));
        goto fail;
    }

    _ulBytesProcessed = 0;
    _ulUnderrunBase   = 0;
    fUnderrun = FALSE;

    ulStreamState = STREAM_STREAMING;
    //Adding the first buffer also starts playback
    if(ulStreamType == STREAM_WAVE_PLAY) {
   	    AddBuffers(TRUE);
        //Must set volume after adding buffers (voices inside sblive driver might not
        //be allocated otherwise (first start) )
        pahw->SetVolume(StreamId, getMixerStreamId(), volume);
    }
    else {
        ulSavedInputGain = pahw->QueryVolume(StreamId, getMixerStreamId());
        pahw->SetInputSrc(StreamId, getMixerStreamId(), inputsrc);
        pahw->SetVolume(StreamId, getMixerStreamId(), inputgain);

	    if(pahw->Start(StreamId) != TRUE) {
            dprintf(("pahw->Start failed!!"));
            goto fail;
        }
    }
//    dprintf(("WAVESTREAM::StartStream %lx", StreamId));
    return NO_ERROR;

fail:
    DebugInt3();
    return ERROR_START_STREAM;
}
//******************************************************************************
//******************************************************************************
ULONG WAVESTREAM::StopStream(PCONTROL_PARM pControl)
{
    if(ulStreamState == STREAM_STOPPED) {
       	dprintf(("WAVESTREAM::StopStream %lx (already stopped)", StreamId));
   	    fUnderrun = FALSE;
   	    pControl->ulTime = GetCurrentTime();
	    return NO_ERROR;
    }
    //silence wave stream before stopping it (removes clicks)
    if(ulStreamType == STREAM_WAVE_PLAY) {
        pahw->SetVolume(StreamId, getMixerStreamId(), 0);
    }
    else {
        pahw->SetVolume(StreamId, getMixerStreamId(), ulSavedInputGain);
    }
    pahw->Stop(StreamId);

    ulStreamState = STREAM_STOPPED;
    fUnderrun = FALSE;
//    dprintf(("WAVESTREAM::StopStream %lx", StreamId));
    ReturnBuffers();
    pControl->ulTime = GetCurrentTime();
    _ulTimeBase = GetCurrentTime();
    return NO_ERROR;
}
//******************************************************************************
//******************************************************************************
ULONG  WAVESTREAM::PauseStream(PCONTROL_PARM pControl)
{
    ULONG endpos;

    pahw->GetPosition(StreamId, &_configinfo, &endpos);
    //silence wave stream before stopping it (removes clicks)
    pahw->SetVolume(StreamId, getMixerStreamId(), 0);
    pahw->Stop(StreamId);

    ulStreamState = STREAM_PAUSED;
    fUnderrun = FALSE;

//    dprintf(("WAVESTREAM::PauseStream %lx", StreamId));
    _vRealignPausedBuffers(endpos);

    _ulBytesProcessed = endpos;
    pControl->ulTime = GetCurrentTime();
    return NO_ERROR;
}
//******************************************************************************
//******************************************************************************
ULONG  WAVESTREAM::ResumeStream(void)
{
#ifdef DEBUG
    dprintf(("WAVESTREAM::ResumeStream %lx", StreamId));
#endif
    return StartStream();
}
//******************************************************************************
//******************************************************************************
BOOL WAVESTREAM::SetProperty(int type, ULONG value, ULONG reserved)
{
#ifdef DEBUG
//    dprintf(("WAVESTREAM::SetProperty"));
#endif

    switch(type) {
    case PROPERTY_VOLUME:
        volume = value; 
        if(ulStreamState == STREAM_STREAMING && ulStreamType == STREAM_WAVE_PLAY) {
            MixerSetWaveVolume(getMixerStreamId(), StreamId, volume);
        }
        break;

    case PROPERTY_INPUTSRC:
        inputsrc = value;
        break;

    case PROPERTY_INPUTGAIN:
        inputgain = value;
        break;

    default:
        return STREAM::SetProperty(type, value, reserved);
    }
    return TRUE;
}
//******************************************************************************
//******************************************************************************
ULONG WAVESTREAM::GetProperty(int type) 
{
#ifdef DEBUG
//    dprintf(("WAVESTREAM::GetProperty"));
#endif
    switch(type) {
    case PROPERTY_FREQUENCY:
       return _configinfo.ulSampleRate;

    case PROPERTY_INPUTSRC:
       return inputsrc;

    case PROPERTY_INPUTGAIN:
       return inputgain;

    default:
       return STREAM::GetProperty(type);
    }
}
//******************************************************************************
//******************************************************************************
WAVESTREAM::WAVESTREAM(ULONG streamtype, LPMCI_AUDIO_INIT pinit, USHORT filesysnum, OSSSTREAMID mixerStreamId):
   STREAM(streamtype, filesysnum, mixerStreamId) 
{
#ifdef DEBUG
//    dprintf(("WAVESTREAM::WAVESTREAM"));
#endif
    // get the pointer to the hardware object
    pahw = (WAVEAUDIO*)GetHardwareDevice(streamtype);

    _fmemset(&_configinfo, 0, sizeof(_configinfo));
    _configinfo.ulSampleRate    = pinit->lSRate;
    _configinfo.ulBitsPerSample = pinit->lBitsPerSRate;
    _configinfo.ulNumChannels   = pinit->sChannels;
    _configinfo.ulDataType      = pinit->sMode;
    _ulBytesProcessed           = 0;
    _ulTimeBase                 = 0;
    _ulUnderrunBase             = 0;
    ulSavedInputGain            = 0;

    _configinfo.pConversionBuffer  = 0;
    pahw->SetupConversion(&_configinfo, pinit->ulOperation, pinit->sMode, pinit->sChannels, pinit->lBitsPerSRate);
   
    if(_configinfo.usConversion != CONVERT_NONE) 
    {//allocate conversion buffer
        PVOID pSelOff;
        LIN   linAddr;

        if(DevHelp_VMAlloc(VMDHA_FIXED, CONVERSION_BUFFER_SIZE, (ULONG)-1L, &linAddr, &pSelOff) != 0) 
        {
             DebugInt3(); //should never happen!!
             _configinfo.usConversion = CONVERT_NONE;
        }
        else _configinfo.pConversionBuffer = linAddr;
    }
    if(_configinfo.fSampleRateConversion == TRUE) 
    {//allocate buffer for sample rate conversion
        PVOID pSelOff;
        LIN   linAddr;

        if(DevHelp_VMAlloc(VMDHA_FIXED, CONVERSION_BUFFER_SIZE, (ULONG)-1L, &linAddr, &pSelOff) != 0) 
        {
             DebugInt3(); //should never happen!!
        }
        else _configinfo.pSRateConvBuffer = linAddr;
    }

    pinit->ulFlags |= pahw->QueryDataFlags(pinit->ulOperation, pinit->sMode, pinit->lBitsPerSRate);

    fUnderrun   = FALSE;
    StreamId    = 0;
    
    if(pahw->Open(current_device, ulStreamType, ulSysFileNum, &StreamId) != TRUE)
    {
        dprintf(("pahw->Open failed!!"));
        DebugInt3();
    }

//    dprintf(("WAVESTREAM stream %lx ctor %lx: rate %d bps %d numchan %d type %x", hstream, StreamId, (USHORT)_configinfo.ulSampleRate, (USHORT)_configinfo.ulBitsPerSample, (USHORT)_configinfo.ulNumChannels, (USHORT)_configinfo.ulNumChannels, (USHORT)_configinfo.ulDataType));
}
//******************************************************************************
//******************************************************************************
WAVESTREAM::~WAVESTREAM()
{
//    dprintf(("WAVESTREAM stream %x dtor %lx", hstream, StreamId));
    if(_configinfo.pConversionBuffer) {
        if(DevHelp_VMFree(_configinfo.pConversionBuffer) != 0) {
            DebugInt3();
        }
    }
    if(_configinfo.pSRateConvBuffer) {
        if(DevHelp_VMFree(_configinfo.pSRateConvBuffer) != 0) {
            DebugInt3();
        }
    }
    if(pahw) {
        if (ulStreamState == STREAM_STREAMING)
           pahw->Stop(StreamId);

    	pahw->Close(StreamId);
        StreamId = 0;
    }
}
//******************************************************************************
//******************************************************************************

